"""
In this package, we provide implementations for basic search operators.

- In :mod:`~moptipy.operators.bitstrings`, you can find operators for elements
  from the :mod:`~moptipy.spaces.bitstrings` space.
- In :mod:`~moptipy.operators.ordered_choices`, you can find operators for
  elements from the :mod:`~moptipy.spaces.ordered_choices` space.
- In :mod:`~moptipy.operators.permutations`, you can find operators for
  elements from the :mod:`~moptipy.spaces.permutations` space.
- In :mod:`~moptipy.operators.vectors`, you can find operators for
  elements from the :mod:`~moptipy.spaces.vectorspace`.
"""
